/*	pcm.c*/
/*	Sound Blaster Pro*/
/*	Joseph Wang*/
/*	11.4.99*/

#include <Drivers.h>
#include <KernelExport.h>
#include <ISA.h>
#include <malloc.h>
#include <stdio.h>
#include "sbpro.h"
#include "pcm.h"

int stall_state=0;

static int32 mode, write_offset, write_size, write_state, write_latency;
static bigtime_t write_time, write_clock;

/*static status_t pcm_open(const char *name, uint32 flags, void **cookie);*/
/*static status_t pcm_close(void *cookie);*/
/*static status_t pcm_free(void *cookie);*/
/*static status_t pcm_control(void *cookie, uint32 op, void *data, size_t len);*/
/*static status_t pcm_read(void *cookie, off_t pos, void *data, size_t *len);*/
/*static status_t pcm_write(void *cookie, off_t pos, const void *data, size_t *len);*/

static status_t pcm_set_params(sound_setup *setup);
static status_t pcm_get_params(sound_setup *setup);
static void set_dma_mode(int32 new_mode, const char *data, int32 len);
/*static int32 pcm_write_inth(void *data);*/

int file;

status_t pcm_open(const char *name, uint32 flags, void **cookie) {
	audio_format *format;
	if(atomic_add(&open_count, 1)!=0) return B_ERROR;
	format=(audio_format *)malloc(sizeof(audio_format));
	format->sample_rate=DEFAULT_SAMPLE_RATE;
	format->channels=DEFAULT_CHANNELS;
	format->format=DEFAULT_FORMAT;
	format->big_endian=DEFAULT_ENDIAN;
	format->buffer_header=DEFAULT_BUFFER_HEADER;
	format->write_buf_size=DEFAULT_WRITE_BUF_SIZE;
	format->rec_buf_size=DEFAULT_REC_BUF_SIZE;
	cookie=(void **)&format;
	mode=MODE_NONE;
	return B_OK;
	}

status_t pcm_close(void *cookie) {
	atomic_add(&open_count, -1);
	return B_OK;
	}

status_t pcm_free(void *cookie) {
	return B_OK;
	}

status_t pcm_control(void *cookie, uint32 op, void *data, size_t len) {
	status_t err;
	audio_buffer_header *header;
	static float rates[]={44100.0, 44100.0, 44100.0, 44100.0};
	switch(op) {
	case B_AUDIO_SET_AUDIO_FORMAT:
		memcpy(cookie, data, min(sizeof(audio_format), len));
		return B_OK;
	case B_AUDIO_GET_AUDIO_FORMAT:
		memcpy(data, cookie, sizeof(audio_format));
		return B_OK;
	case B_AUDIO_GET_PREFERRED_SAMPLE_RATES:
		memcpy(data, rates, sizeof(rates));
		return B_OK;
	case SOUND_GET_PARAMS:
		return pcm_get_params((sound_setup *)data);
	case SOUND_SET_PARAMS:
		return pcm_set_params((sound_setup *)data);
/*	case SOUND_SET_CAPTURE_COMPLETION_SEM:*/
	case SOUND_SET_PLAYBACK_COMPLETION_SEM:
		write_sem=*((sem_id *)data);
		return B_OK;
	case SOUND_UNSAFE_WRITE: 
		header=(audio_buffer_header *)data;
		len=header[0].reserved_1-sizeof(header[0]);
		err=pcm_write(cookie, 0, (void *)&header[1], &len);
		header[0].sample_clock=(write_clock*1000)/1764;
		header[0].time=header[0].sample_clock;
		return release_sem(write_sem);
	case 10012:
	case 10013:
		*(int32*)data = B_PAGE_SIZE;
		return B_OK;
		}
	return EINVAL;
	}

status_t pcm_read(void *cookie, off_t pos, void *data, size_t *len) {
	return B_OK;
	}

status_t pcm_write(void *cookie, off_t pos, const void *data, size_t *len) {
	int32 i, togo=*len, silent;
	uchar *raw=(uchar *)data;
	audio_format *format=(audio_format *)cookie;
/*i=*len;*/
/*file=open("/boot/sbpro", O_CREAT|O_RDWR|O_APPEND);*/
/*write(file, &i, 4);*/
/*close(file);*/
	if (acquire_sem_etc(write_sem, 1, B_CAN_INTERRUPT | B_TIMEOUT, 1000000LL) < B_OK) return EIO;
/*	decimate 16-bit/44.1k to 8-bit/22.05k*/
	if(togo&3<3) togo-=4;
	togo>>=2;
	if(togo<=0) {
		release_sem(write_sem);
		return B_OK;
		}
	silent=1;
	for(i=0; i<togo; i+=2) {
		raw[i]=raw[i*4+3]+0x80;
		raw[i+1]=raw[i*4+1]+0x80;
		if(raw[i+1]!=0x80 || raw[i]!=0x80) silent=0;
		}
	if(silent) {
		for(i=0; i<togo; i+=2) {
			raw[i+1]=0;
			}
		}
//	raw[1]=0;
//	memset(raw+togo-256, 0, togo);
	set_dma_mode(MODE_WRITE_8BIT_AUTO, data, togo);
	for(;togo>0;) {
		if(write_size>=2*write_latency) {
			snooze(2000);
			continue;
			}
		if(write_state==0) {
			snooze(2000);
			continue;
			}
		i=min(togo, write_latency);
		if(write_offset>=write_limit) write_offset=0;
		memcpy(write_buffer.data+write_offset, raw, i);
/*		write silence*/
		if(i<write_latency) {
			memset(write_buffer.data+write_offset+i, 128, write_latency-i);
			}
		togo-=i;
		raw+=i;
		write_offset+=write_latency;
		write_size+=write_latency;
		write_state=0;
		}
	acquire_sl();
	write_data(SBPRO_EXIT_DMA);
	release_sl();
	write_state=0;
	write_clock+=(*len)&0xfc;
	release_sem_etc(write_sem, 1, B_DO_NOT_RESCHEDULE);
	return B_OK;
	}

/*support functions*/

void set_dma_mode(int32 new_mode, const char *data, int32 len) {
	int32 size;
	physical_entry table;
	if(new_mode==mode) {
/*		unlock dma channel*/
		switch(mode) {
			physical_entry table;
		case MODE_WRITE_8BIT_AUTO:
			write_state=0;
			acquire_sl();
			write_data(SBPRO_PLAYBACK_8_BIT);
			release_sl();
			}
		return;
		}
	mode=new_mode;
	switch(mode) {
		case MODE_WRITE_8BIT_AUTO: {
			physical_entry table;
			get_memory_map(write_buffer.data, write_buffer.size, &table, 1);
			isa_info->start_isa_dma(dma8, table.address, table.size, DMA_IO_WRITE, 0);
			memset(write_buffer.data, 128, write_latency);
			write_limit=write_buffer.size;
			write_time=system_time();
			write_clock=0;
			write_state=0;
			write_size=0;
/*			write_latency=min(len, write_limit);*/
			write_latency=min(len, DEFAULT_WRITE_BUF_SIZE>>2);
			write_offset=0;//write_latency;
			acquire_sl();
/*			stereo*/
			write_reg(0xe, 0x23);
			write_data(0x42);
			write_data(SBPRO_TIME_CONSTANT);
			write_data(233);
/*			filter constant*/
			write_data(0x41);
			write_data(233);
			write_data(SBPRO_TRANSFER_SIZE);
			write_data((write_latency-1)&0xff);
			write_data((write_latency-1)>>8);
			write_data(SBPRO_PLAYBACK_8_BIT);
			write_data(SBPRO_SPEAKER_ENABLE);
			release_sl();
			}
		}
/*	release_sl();*/
	}

int32 pcm_write_inth(void *data) {
	write_size-=write_latency;
	acquire_sl();
	read_io(SBPRO_READ_BUFFER_STATUS);
	write_state=1;
	release_sl();
	return B_INVOKE_SCHEDULER;
	}

status_t pcm_set_params(sound_setup *sound) {
	int value;
/*			no controls for channel.adc_gain*/
/*			set channel.mic levels, 1bit->2bit*/
/*			if(sound->left.mic_gain_enable*/
/*					||sound->right.mic_gain_enable) {*/
/*				write_reg(SBPRO_MIC, 4);*/
/*				}*/
/*			else write_reg(SBPRO_MIC, 0);*/
/*			aux1 maps to CD-ROM, 5bit->4bit*/
	value=0;
	if(!sound->left.aux1_mix_mute)
			value+=224-((sound->left.aux1_mix_gain<<3)&0xe0);
	if(!sound->right.aux1_mix_mute)
			value+=14-((sound->right.aux1_mix_gain>>1)&0x0e);
	acquire_sl();
	write_reg(SBPRO_CD, value);
	release_sl();
/*			aux2 maps to MIDI, 5->4*/
	value=0;
	if(!sound->left.aux2_mix_mute)
			value+=224-((sound->left.aux2_mix_gain<<3)&0xe0);
	if(!sound->right.aux2_mix_mute)
			value+=14-((sound->right.aux2_mix_gain>>1)&0x0e);
	acquire_sl();
	write_reg(SBPRO_MIDI, value);
	release_sl();
/*			set line levels, 5->4*/
	value=0;
	if(!sound->left.line_mix_mute)
			value+=224-((sound->left.line_mix_gain<<3)&0xe0);
	if(!sound->right.line_mix_mute)
			value+=14-((sound->right.line_mix_gain>>1)&0x0e);
	acquire_sl();
	write_reg(SBPRO_LINE, value);
	release_sl();
/*			dac maps to voice, 6->4*/
	value=0;
	if(!sound->left.dac_mute)
			value+=224-((sound->left.dac_attn<<2)&0xe0);
	if(!sound->right.dac_mute)
			value+=14-((sound->right.dac_attn>>2)&0x0e);
	acquire_sl();
	write_reg(SBPRO_VOICE, value);
	release_sl();
/*			only support 22.1kHz*/
/*			only support 8 bit stereo playback*/
/*			only support 8 bit stereo capture*/
/*			no control for dither_enable*/
/*			no control for loop_attn*/
/*			no control for loop back*/
/*			no control for output boosk*/
/*			no control for high pass filter*/
/*			no control for speaker gain*/
/*			un/mute speaker*/
/*			if(sound->mono_mute) hw_codec_write_byte(SB16_SPEAKER_DISABLE);*/
/*			else hw_codec_write_byte(SB16_SPEAKER_ENABLE);*/
	return B_OK;
	}

status_t pcm_get_params(sound_setup *sound) {
	int value, value2;
/*	acquire_sl();*/
/*	release_sl();*/
/*return B_OK;*/
/*			no controls for channel.adc_gain*/
	sound->left.adc_gain=sound->right.adc_gain = 15;
/*			set channel.mic levels, 1bit->2bit*/
	sound->left.mic_gain_enable=0;
	sound->right.mic_gain_enable=0;
/*			if(read_reg(SBPRO_MIC)&6) {*/
/*				sound->left.mic_gain_enable=1;*/
/*				sound->right.mic_gain_enable=1;*/
/*				}*/
/*			else {*/
/*				sound->left.mic_gain_enable=0;*/
/*				sound->right.mic_gain_enable=0;*/
/*				}*/
/*			CD-ROM maps to aux1, 4bit->5bit, no mute control*/
	value=value2=read_reg(SBPRO_CD);
	value=30-((value>>3)&30);
	value2=30-((value2<<1)&30);
	sound->left.aux1_mix_gain=value;
	sound->right.aux1_mix_gain=value2;
	sound->left.aux1_mix_mute=sound->right.aux1_mix_mute=0;
/*			MIDI maps to aux2, 4->5*/
	value=value2=read_reg(SBPRO_MIDI);
	value=30-((value>>3)&30);
	value2=30-((value2<<1)&30);
	sound->left.aux2_mix_gain=value;
	sound->right.aux2_mix_gain=value2;
	sound->left.aux2_mix_mute=sound->right.aux2_mix_mute=0;
/*			set line levels, 4->5*/
	value=value2=read_reg(SBPRO_LINE);
	value=30-((value>>3)&30);
	value2=30-((value2<<1)&30);
	sound->left.line_mix_gain=value;
	sound->right.line_mix_gain=value2;
	sound->left.line_mix_mute=sound->right.line_mix_mute=0;
/*			dac maps to voice, 6->4*/
	value=value2=read_reg(SBPRO_VOICE);
	value=56-((value>>2)&56);
	value2=56-((value2<<2)&56);
	sound->left.dac_attn=value;
	sound->right.dac_attn=value2;
	sound->left.dac_mute=sound->right.dac_mute=0;

/*			only support 22.05kHz*/
	sound->sample_rate = kHz_22_05;
/*			only support 8 bit stereo playback*/
	sound->playback_format = linear_16bit_little_endian_stereo;
/*sound->playback_format = linear_8bit_unsigned_stereo;*/
/*			only support 8 bit stereo capture*/
	sound->capture_format = linear_8bit_unsigned_stereo;
/*			no control for dither_enable*/
	sound->dither_enable = 0;
/*			no control for loop_attn*/
	sound->loop_attn = 64;
/*			no control for loop back*/
	sound->loop_enable = 0;
/*			no control for output boost*/
	sound->output_boost = 0;
/*			no control for high pass filter*/
	sound->highpass_enable = 0;
/*			no control for speaker gain*/
	sound->mono_gain = 64;
/*			un/mute speaker*/
	sound->mono_mute = 0;
/*			if(sound->mono_mute) hw_codec_write_byte(SB16_SPEAKER_DISABLE);*/
/*			else hw_codec_write_byte(SB16_SPEAKER_ENABLE);*/
	return B_OK;
	}
